import json
import os
import tempfile
from typing import List, Optional

import pandas as pd
from loguru import logger
from tqdm import tqdm

from skit_fixdf import constants as const
from skit_fixdf.fix.df import (
    duplicate_columns,
    fix_datetime,
    min_json_whitespace,
    read_df,
    remove_columns,
    remove_rows,
    rename_columns,
    reorder_cols,
)


def add_entities_col(df: pd.DataFrame) -> pd.DataFrame:
    """
    Create entities column from prediction column.

    :param df: Dataframe.
    :return: Dataframe.
    """
    if const.PREDICTION not in df.columns:
        return df

    df[const.ENTITIES] = None

    for i, value in tqdm(
        df.prediction.iteritems(),
        total=len(df),
        desc="Extract predicted entities into a separate column.",
    ):
        entities = []
        if not isinstance(value, str):
            df.at[i, const.ENTITIES] = json.dumps(entities, ensure_ascii=False)
            continue
        prediction = json.loads(value)
        intents = prediction.get(const.INTENTS, [])
        for intent in intents:
            for slot in intent.get(const.SLOTS, []):
                for entity_value in slot.get(const.VALUES, []):
                    entity = {
                        const.TEXT: entity_value.get(const.BODY),
                        const.TYPE: entity_value.get(const.TYPE)
                        or entity_value.get(const.ENTITY_TYPE)
                        or entity_value.get(const.TRANSCRIPTION),
                        const.SCORE: entity_value.get(const.SCORE, 0),
                        const.VALUE: entity_value.get(const.VALUE),
                    }
                    entities.append(entity)
        df.at[i, const.ENTITIES] = json.dumps(entities, ensure_ascii=False)
    return df


def add_state_transitions_col(
    df: pd.DataFrame, transition_token: str = "->"
) -> pd.DataFrame:
    """
    Create state_transitions_col from states.

    :param df: Source df with states
    :type df: pd.DataFrame
    :return: df with state transitions
    :rtype: pd.DataFrame
    """
    df[const.STATE_TRANSITIONS] = None

    if const.CALL_UUID not in df.columns:
        return df

    if const.STATE not in df.columns:
        return df

    unique_call_ids = df[const.CALL_UUID].unique()
    for call_id in tqdm(
        unique_call_ids, desc="Extract state transitions into a separate column."
    ):
        call_slice = df[df[const.CALL_UUID] == call_id]
        states = call_slice[const.STATE].values
        df.loc[call_slice.index, const.STATE_TRANSITIONS] = transition_token.join(
            states
        )
    return df


def add_asr_provider_col(df: pd.DataFrame) -> pd.DataFrame:
    """
    Add asr_provider column to dataframe.

    :param df: Source df with metadata column.
    :type df: pd.DataFrame
    :return: df with asr_provider column
    :rtype: pd.DataFrame
    """
    df[const.ASR_PROVIDER] = None

    if const.METADATA not in df.columns:
        return df

    for i, value in tqdm(
        df.metadata.iteritems(),
        total=len(df),
        desc="Extract asr_provider into a separate column.",
    ):
        value = json.loads(value)
        df.at[i, const.ASR_PROVIDER] = value.get(const.ASR_PROVIDER)
    return df


def add_utterances_col(df: pd.DataFrame) -> pd.DataFrame:
    """
    Add utterances column to dataframe.

    :param df: Source df with debug_metadata column.
    :type df: pd.DataFrame
    :return: df with utterances column
    :rtype: pd.DataFrame
    """
    df[const.UTTERANCES] = None

    if const.DEBUG_METADATA not in df.columns:
        return df

    for i, value in tqdm(
        df.debug_metadata.iteritems(),
        total=len(df),
        desc="Extract utterances into a separate column.",
    ):
        value = json.loads(value)
        utterances = value.get(const.PLUTE_REQUEST, {}).get(const.ALTERNATIVES, [])
        df.at[i, const.UTTERANCES] = json.dumps(utterances, ensure_ascii=False)
    return df


def prepare_call_df_for_tagging(
    df_path: str, output_path: Optional[str] = None, delete_file: bool = False
) -> str:
    """
    Prepare call df for tagging.

    :param df_path: Path where the calls df exists.
    :return: Dataframe.
    :rtype: pd.DataFrame
    """
    df = read_df(df_path)
    logger.debug(f"Read df from {df_path=} of shape={df.shape}")

    df = remove_rows(
        df, [const.SUB_TYPE], const.SUPPORTED_SUB_TYPES(df[const.SUB_TYPE].unique())
    )
    logger.debug(f"After removing selected rows {df.shape}")

    df = duplicate_columns(df, [(const.PREDICTED_INTENT, const.INTENT)])
    logger.debug(f"After duplicating columns {df.shape}")

    df = fix_datetime(df, [const.CREATED_AT])
    logger.debug(f"Fixed datetime {df.shape}")

    df = rename_columns(
        df,
        {
            const.CREATED_AT: const.REFTIME,
            const.VIVA_CALL_DURATION: const.CALL_DURATION,
        },
    )
    df = add_entities_col(df)
    df = add_state_transitions_col(df)
    df = add_asr_provider_col(df)
    df = add_utterances_col(df)
    df = min_json_whitespace(
        df,
        [
            const.METADATA,
            const.ENTITIES,
            const.UTTERANCES,
            const.PREDICTION,
        ],
    )
    df = remove_columns(df, df.columns.difference(const.REQUIRED_COLS_FOR_CALLS))
    df = reorder_cols(df, const.REQUIRED_COLS_FOR_CALLS)

    if not output_path:
        input_file_dir, input_file_name = os.path.split(df_path)
        input_file_name, input_file_ext = os.path.splitext(input_file_name)
        _, output_path = tempfile.mkstemp(
            dir=input_file_dir, prefix=f"{input_file_name}-", suffix=input_file_ext
        )

    df.to_csv(output_path, index=False)
    if delete_file:
        os.remove(df_path)
    return output_path



DEBUG_METADATA = "debug_metadata"
STATE = "state"
TYPE = "type"
SUB_TYPE = "sub_type"
AUDIO = "audio"
CONVERSATION_UUID = "conversation_uuid"
CALL_UUID = "call_uuid"
CALL_AUDIO = "call_audio"
CREATED_AT = "created_at"
END_STATE = "end_state"
FLOW_VERSION = "flow_version"
IS_REPORTED = "is_reported"
LANGUAGE_CODE = "language_code"
VIRTUAL_NUMBER = "virtual_number"
CALL_DURATION = "call_duration"
BOT = "bot"
CURRENT_INTENT = "current_intent"
AUDIO_URL = "audio_url"
PREDICTED_INTENT = "predicted_intent"
PREDICTION = "prediction"
USER = "user"
VIVA_CALL_DURATION = "viva_call_duration"

DATA = "data"
DATA_ID = "data_id"
ID = "id"
NAME = "name"
INTENT = "intent"
INTENTS = f"{INTENT}s"
SLOTS = "slots"
VALUES = "values"
ENTITIES = "entities"
TEXT = "text"
BODY = "body"
TYPE = "type"
ENTITY_TYPE = "entity_type"
TRANSCRIPTION = "transcription"
SCORE = "score"
VALUE = "value"
STATE_TRANSITIONS = "state_transitions"
METADATA = "metadata"
ASR_PROVIDER = "asr_provider"
UTTERANCES = "utterances"
PLUTE_REQUEST = "plute_request"
ALTERNATIVES = "alternatives"
REFTIME = "reftime"
TAG = "tag"
TYPE = "type"
RAW = "raw"
IS_GOLD = "is_gold"


REQUIRED_COLS_FOR_CALLS = [
    # Identifiers
    CONVERSATION_UUID,
    CALL_UUID,
    REFTIME,
    # audio url
    CALL_AUDIO,
    AUDIO_URL,
    # flow params
    FLOW_VERSION,
    STATE,
    END_STATE,
    IS_REPORTED,
    LANGUAGE_CODE,
    VIRTUAL_NUMBER,
    STATE_TRANSITIONS,
    CALL_DURATION,
    # slu metadata
    UTTERANCES,
    CURRENT_INTENT,
    PREDICTION,
    INTENT,
    ENTITIES,
    BOT,
    USER,
]

LABEL_COLS_FROM_METADATA = [
    CALL_UUID,
    CONVERSATION_UUID,
    STATE,
    ALTERNATIVES,
    REFTIME,
    PREDICTION,
    INTENT,
    ENTITIES,
    LANGUAGE_CODE,
]

SUPPORTED_SUB_TYPES = lambda l: [v for v in l if v.lower() != AUDIO]

LOCALE = "locale"
TIMEZONE = "tz"
DIMS = "dims"
LATENT = "latent"
DATE = "date"
TIME = "time"
DATETIME = "datetime"
NUMBER = "number"
DURATION = "duration"
SUPPORTED_ENTITIES = [DATE, TIME, DATETIME, NUMBER, DURATION]
DATASET_TYPE__INTENT = "intent"
DATASET_TYPE__ENTITY = "entity"
DATASET_TYPE__TRANSCRIPTION = "transcription"
SUPPORTED_DATASET_TYPES = [
    DATASET_TYPE__TRANSCRIPTION,
    DATASET_TYPE__INTENT,
    DATASET_TYPE__ENTITY,
]
TO = "to"
FROM = "from"
DIM = "dim"
DURATION = "duration"
SECOND = "second"
NORMALIZED = "normalized"
DUCKLING_SERVICE_URL = "DUCKLING_SERVICE_URL"
UNIT = "unit"
CALLS = "calls"
LABELS = "labels"
INNER = "inner"



"""
Function to convert arbitrary datetime strings to ISO 1806 format.
"""

from datetime import datetime


def to_datetime(d: str) -> datetime:
    """
    Convert an arbitrary date string to ISO format.

    :param d: A date string.
    :type d: str
    :raises ValueError: If the date string can't be parsed.
    :return: The date string in ISO format.
    :rtype: datetime
    """
    if not isinstance(d, str):
        raise TypeError(f"Expected a string, got {type(d)}")

    time_fns = [
        datetime.fromisoformat,
        lambda date_string: datetime.strptime(
            date_string, "%Y-%m-%d %H:%M:%S.%f %z %Z"
        ),
        lambda date_string: datetime.strptime(date_string, "%Y-%m-%dT%H:%M:%SZ"),
        lambda date_string: datetime.strptime(date_string, "%Y-%m-%dT%H:%M:%S.%f%z"),
    ]

    for time_fn in time_fns:
        try:
            return time_fn(d)
        except ValueError:
            continue
    raise ValueError(f"Could not parse date string: {d}")



"""
CLI for formatting dataframes.
"""
import argparse
import os
import sys

from skit_fixdf import constants as const
from skit_fixdf import utils
from skit_fixdf.fix import calls_df, labels_df


def build_calls_cmd(parser: argparse.ArgumentParser):
    parser.add_argument("--input", "-i", help="Input csv file.")
    parser.add_argument("--output", "-o", help="Output csv file.")


def build_labels_cmd(parser: argparse.ArgumentParser):
    parser.add_argument("--input", "-i", help="Input csv file.")
    parser.add_argument("--output", "-o", help="Output csv file.")
    parser.add_argument(
        "--dataset-type",
        "-dt",
        required=True,
        help="Dataset type.",
        choices=const.SUPPORTED_DATASET_TYPES,
    )
    parser.add_argument("--duckling-url", help="Duckling host for entity parsing.")


def build_parser():
    parser = argparse.ArgumentParser(
        "Formatting/Patching dataframes obtained via skit-calls and skit-df."
    )
    parser.add_argument(
        "-v", "--verbose", action="count", default=0, help="Increase verbosity."
    )
    source = parser.add_subparsers(dest="source", help="Source of dataframe.")
    calls_cmd = source.add_parser(
        "calls", help="Untagged Dataframe obtained via skit-calls."
    )
    build_calls_cmd(calls_cmd)
    labels_cmd = source.add_parser(
        "labels", help="Tagged Dataframe obtained via skit-df."
    )
    build_labels_cmd(labels_cmd)
    return parser


def main():
    args = build_parser().parse_args()
    utils.configure_logger(args.verbose)

    if args.input is None:
        is_pipe = not os.isatty(sys.stdin.fileno())
        if is_pipe:
            args.input = sys.stdin.readline().strip()
        else:
            raise argparse.ArgumentTypeError(
                "Expected to receive --input=<file> or its valued piped in."
            )

    if args.source == const.CALLS:
        output = calls_df.prepare_call_df_for_tagging(args.input, args.output)
        print(output)
    elif args.source == const.LABELS:
        output = labels_df.prepare_tagged_df_for_training(
            args.input, args.dataset_type, args.output, duckling_url=args.duckling_url
        )
        print(output)
    else:
        raise ValueError(f"Unknown source {args.source=}.")



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['skit_fixdf', 'skit_fixdf.fix']

package_data = \
{'': ['*']}

install_requires = \
['loguru>=0.5.3,<0.6.0', 'pandas>=1.3.5,<2.0.0', 'tqdm>=4.62.3,<5.0.0']

entry_points = \
{'console_scripts': ['skit-fixdf = skit_fixdf.cli:main']}

setup_kwargs = {
    'name': 'skit-fixdf',
    'version': '0.1.8',
    'description': "A library to format datasets so that you don't have to.",
    'long_description': None,
    'author': 'ltbringer',
    'author_email': 'amresh.venugopal@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/skit-ai/skit-fixdf',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<3.10',
}


setup(**setup_kwargs)



import json
import ast
from typing import Dict, List, Tuple

import pandas as pd

from skit_fixdf import constants as const
from skit_fixdf.fix.datetime import to_datetime


def ensure_columns(fn):
    def wrapper(*args, **kwargs):
        columns = kwargs.get(const.COLUMNS)
        df = args[0]

        if not columns:
            return fn(*args, **kwargs)

        if not isinstance(columns, (list, dict)):
            return fn(*args, **kwargs)

        if not isinstance(df, pd.DataFrame):
            return fn(*args, **kwargs)
        
        if isinstance(columns, dict):
            kwargs[const.COLUMNS] = {k: v for k, v in columns.items() if k in df.columns}
            return fn(*args, **kwargs)

        if all(isinstance(column, tuple) for column in columns):
            kwargs[const.COLUMNS] = filter(lambda col_pairs: col_pairs[0] in df.columns, columns)
        else:
            kwargs[const.COLUMNS] = filter(lambda column: column in df.columns, columns)
        
        return fn(*args, **kwargs)
    return wrapper


def read_df(df_path: str) -> pd.DataFrame:
    """
    Reads a dataframe from a file.
    :param df_path: Path to the dataframe.
    :return: Dataframe.
    """
    df = pd.read_csv(df_path)
    return df


def rejson(data: str) -> str:
    """
    Re-encode a JSON string.

    :param data: A JSON string.
    :type data: str
    :return: The re-encoded JSON string.
    :rtype: str
    """
    if not isinstance(data, (str, dict)):
        return data

    dict_deserializers = [json.loads, ast.literal_eval]

    if isinstance(data, dict):
        return json.dumps(data, ensure_ascii=False)

    for deserz in dict_deserializers:
        try:
            data = deserz(data)
            return json.dumps(data, ensure_ascii=False)
        except (json.JSONDecodeError, ValueError):
            continue

    raise ValueError(f"Could not decode {data} as json")


@ensure_columns
def json_loads(df: pd.DataFrame, columns: List[str]) -> pd.DataFrame:
    for column in columns:
        df[column] = df[column].apply(lambda el: json.loads(el), axis=1)
    return df


def are_invalid_utterances(utterances: List[dict]) -> bool:
    return all(isinstance(utterance, dict) for utterance in utterances) and not pd.isna(utterances)


def fix_utterances(df: pd.DataFrame, column: str) -> pd.DataFrame:
    df[column] = df[column].apply(lambda el: [el] if are_invalid_utterances(el) else el)
    return df


@ensure_columns
def reorder_cols(df: pd.DataFrame, columns: List[str]) -> pd.DataFrame:
    """
    Reorder columns.

    :param df: Dataframe.
    :return: Dataframe.
    """
    return df[columns]


@ensure_columns
def remove_columns(df: pd.DataFrame, columns: List[str]) -> pd.DataFrame:
    """
    Removes columns from a dataframe.
    :param df: Dataframe.
    :param columns: List of columns to remove.
    :return: Dataframe.
    """
    df.drop(columns, axis=1, inplace=True)
    return df


@ensure_columns
def remove_rows(
    df: pd.DataFrame, columns: List[str], values: List[str]
) -> pd.DataFrame:
    """
    Removes rows from a dataframe.

    :param df: Dataframe.
    :param columns: List of columns to check.
    :param values: List of values to remove.
    :return: Dataframe.
    """
    for column in columns:
        df = df[~df[column].isin(values)]
    return df


@ensure_columns
def duplicate_columns(df: pd.DataFrame, column_pairs: List[Tuple[str]]) -> pd.DataFrame:
    """
    Duplicates columns in a dataframe.

    We take a data-frame: df and column-pairs: [(a, b), ...].
    We duplicate column a (should exist) into b (to be created).

    :param df: Dataframe.
    :param column_pairs: List of column pairs (a, b).
    :return: Dataframe.
    """
    for (column_a, column_b) in column_pairs:
        df[column_b] = df[column_a]
    return df


def rename_columns(df: pd.DataFrame, column_map: Dict[str, str]) -> pd.DataFrame:
    """
    Renames columns in a dataframe.
    :param df: Dataframe.
    :param columns: Dictionary of columns to rename.
    :return: Dataframe.
    """
    df.rename(columns=column_map, inplace=True)
    return df


@ensure_columns
def fix_datetime(df: pd.DataFrame, columns: List[str]) -> pd.DataFrame:
    """
    Fixes datetime columns in a dataframe.
    :param df: Dataframe.
    :return: Dataframe.
    """
    for column in columns:
        df[column] = df[column].apply(lambda dt: to_datetime(dt).isoformat())
    return df


@ensure_columns
def min_json_whitespace(df: pd.DataFrame, columns: List[str]) -> pd.DataFrame:
    """
    Converts JSON columns to have minimum whitespace.

    :param df: Dataframe.
    :return: Dataframe.
    """
    for column in columns:
        df[column] = df[column].apply(rejson)
    return df



import json
import os
import re
import tempfile
from datetime import datetime
from typing import List, Optional

import pandas as pd
import pytz
import requests
from tqdm import tqdm

from skit_fixdf import constants as const
from skit_fixdf import fix


def make_intent_label_df(df: pd.DataFrame) -> pd.DataFrame:
    """
    Make intent label df.
    :param df: Dataframe.
    :return: Dataframe.
    :rtype: pd.DataFrame
    """
    df = df[[const.ID, const.TAG]].copy()
    for i, row in tqdm(df.iterrows(), total=len(df), desc="Make intent label df."):
        tag_objects = json.loads(row[const.TAG])
        tags = [tag[const.TYPE] for tag in tag_objects]
        if tags:
            head, _ = tags[0], tags[1:]
            df.at[i, const.INTENT] = head
    return df[[const.ID, const.INTENT]]


# ==========================================================================================
# TODO: This section should be deprecated.
# We are adding it here temporarily because of problems with the tagging interface.
# Once True values are available on the interface, we will let this go.
# Not focusing on code quality for the same reason.

# @deprecate
def detect_language(line):
    line = re.sub(r"[^a-zA-Z\u0900-\u097F]+", " ", line)
    maxchar = max(line)
    if "\u0900" <= maxchar <= "\u097f":
        return "hi"
    elif "a" <= maxchar <= "z":
        return "en"
    return None


# @deprecate
def get_entities_from_duckling(
    duckling_url: str,
    text: str,
    lang: str,
    reftime: int,
    type_: str,
    duckling_token: Optional[str] = None,
):
    if not text or not lang:
        return None

    headers = {"Authorization": f"Bearer {duckling_token}"} if duckling_token else {}

    timezone = "Asia/Kolkata"  # Should be UTC but this is needed internally.
    # Not fixing because this function isn't required.

    payload = {
        const.TEXT: text,
        const.LOCALE: f"{lang}_IN",  # this piece is otherwise problematic (hardcoded "IN")
        # but we want to let go of this function altogether.
        const.TIMEZONE: timezone,
        const.DIMS: json.dumps([type_]),
        const.REFTIME: reftime,
        const.LATENT: False,
    }

    timezone = pytz.timezone(timezone)
    value = None
    response = requests.post(
        duckling_url,
        data=payload,
        headers=headers,
        timeout=5,
    )

    if response.status_code == 200:
        entities_list = response.json()
        if entities_list:
            entity = entities_list[0]
            value_store = entity.get(const.VALUE, {})
            if const.VALUE in value_store:
                value = value_store[const.VALUE]
            elif const.FROM in value_store and const.TO in value_store:
                value = {
                    const.FROM: value_store.get(const.FROM),
                    const.TO: value_store.get(const.TO),
                }
            elif const.FROM in value_store:
                value = {const.FROM: value_store.get(const.FROM)}
            elif const.TO in value_store:
                value = {const.TO: value_store.get(const.TO)}

            if entity[const.DIM] == const.DURATION:
                normalized_value = entity.get(const.VALUE, {}).get(const.NORMALIZED, {})
                if normalized_value.get(const.UNIT) == const.SECOND:
                    value = reftime + normalized_value.get(const.VALUE)
                    try:
                        value = datetime.fromtimestamp(value / 1000, timezone)
                        value = value.isoformat()
                    except ValueError:
                        value = None
    return value


# @deprecate
def make_entity_label_df(
    df: pd.DataFrame, duckling_url: str, duckling_token: Optional[str] = None
) -> pd.DataFrame:
    """
    Make entity label df.
    :param df: Dataframe.
    :return: Dataframe.
    :rtype: pd.DataFrame
    """
    df = df[[const.ID, const.TAG, const.LANGUAGE_CODE, const.REFTIME]].copy()
    for i, row in tqdm(
        df.iterrows(), total=len(df), desc="making duckling hits to get entity values."
    ):
        tags = json.loads(row[const.TAG])
        evaluated_tags = []
        for tag in tags:
            if not tag[const.TYPE] or not tag[const.TEXT]:
                continue
            type_ = None
            value = None
            if tag[const.TYPE].lower() in const.SUPPORTED_ENTITIES:
                type_ = tag[const.TYPE].lower()
                reftime = row[const.REFTIME]
                timestamp = datetime.fromisoformat(reftime).timestamp()
                unix_epoch = int(timestamp * 1000)
                language_code = (
                    row[const.LANGUAGE_CODE]
                    if not pd.isna(row[const.LANGUAGE_CODE])
                    else detect_language(tag[const.TEXT])
                )
                value = get_entities_from_duckling(
                    duckling_url,
                    tag[const.TEXT],
                    language_code,
                    unix_epoch,
                    tag[const.TYPE],
                    duckling_token=duckling_token,
                )
            else:
                type_ = tag[const.TYPE]
                value = tag[const.TEXT]

            if type_ and value:
                evaluated_tags.append(
                    {
                        const.TEXT: tag[const.TEXT],
                        const.TYPE: type_,
                        const.SCORE: tag.get(const.SCORE, 0),
                        const.VALUE: value,
                    }
                )
        df.at[i, const.ENTITIES] = json.dumps(evaluated_tags, ensure_ascii=False)
    return df[[const.ID, const.ENTITIES]]


# The above section is set for deprecation.
# ==========================================================================================


def make_transcription_label_df(df: pd.DataFrame) -> pd.DataFrame:
    """
    Make transcription label df.

    :param df: Dataframe.
    :return: Dataframe.
    :rtype: pd.DataFrame
    """
    df = df[[const.ID, const.TAG]].copy()
    df[const.TRANSCRIPTION] = df[const.TAG].apply(
        lambda tag: json.loads(tag).get(const.TEXT) if isinstance(tag, str) else None
    )
    return df[[const.ID, const.TRANSCRIPTION]]


def add_columns_from_metadata(
    df: pd.DataFrame, columns: List[str], key: str = const.DATA, fn: callable = None
) -> pd.DataFrame:
    """
    Extract value of key in separate columns.

    :param df: Dataframe.
    :return: Dataframe.
    :rtype: pd.DataFrame
    """
    if key not in df.columns:
        return df
    for i, row in tqdm(
        df.iterrows(), total=len(df), desc=f"Add values from {fn=} {key=}."
    ):
        raw_data = fn(row[key]) if fn else row[key]
        for col in columns:
            if isinstance(col, tuple):
                from_col, to_col = col
            else:
                from_col = to_col = col

            if isinstance(raw_data, dict):
                value = raw_data.get(from_col)
                if isinstance(value, (dict, list)):
                    value = json.dumps(value, ensure_ascii=False, indent=0)
            else:
                value = None
            df.at[i, to_col] = value
    return df


def prepare_tagged_df_for_training(
    df_path: str,
    dataset_type: str,
    output_path: Optional[str] = None,
    duckling_url: Optional[str] = None,
    duckling_token: Optional[str] = None,
    delete_file: bool = False,
) -> str:
    """
    Prepare tagged df for training on dialogy.

    :param df_path: Path where tagged dataset exists.
    :type df_path: str
    :return: Dialogy compatible dataset.
    :rtype: pd.DataFrame
    """
    if dataset_type == const.DATASET_TYPE__ENTITY and not duckling_url:
        raise ValueError(
            f"duckling url is required for processing {dataset_type} dataset."
            " Refer to https://github.com/skit-ai/skit-auth for more details."
        )

    df = fix.df.read_df(df_path)
    df = add_columns_from_metadata(
        df,
        const.LABEL_COLS_FROM_METADATA,
        fn=lambda x: json.loads(x),
    )
    df = add_columns_from_metadata(
        df,
        [(const.NAME, const.INTENT)],
        key=const.PREDICTION,
        fn=lambda x: json.loads(x),
    )
    df = fix.df.json_loads(df, [const.ALTERNATIVES])
    df = fix.df.fix_utterances(df, const.ALTERNATIVES)
    df = fix.df.fix_datetime(df, [const.REFTIME])
    df = fix.df.rename_columns(df, {const.DATA_ID: const.ID, const.ALTERNATIVES: const.UTTERANCES})

    if dataset_type == const.DATASET_TYPE__INTENT:
        label_df = make_intent_label_df(df)
    elif dataset_type == const.DATASET_TYPE__TRANSCRIPTION:
        label_df = make_transcription_label_df(df)
    elif dataset_type == const.DATASET_TYPE__ENTITY:
        label_df = make_entity_label_df(df, duckling_url, duckling_token=duckling_token)
    else:
        raise ValueError(f"Given {dataset_type=} not supported.")

    if label_df.empty:
        raise ValueError(f"No labels found for {dataset_type=}.")

    df = fix.df.remove_columns(df, [const.DATA, const.TAG, const.IS_GOLD])
    df = fix.df.min_json_whitespace(df, [
        const.UTTERANCES,
        const.PREDICTION,
        const.ENTITIES,
        const.TRANSCRIPTION
    ])

    input_file_dir, input_file_name = os.path.split(df_path)
    input_file_name, input_file_ext = os.path.splitext(input_file_name)

    if not output_path:
        _, output_path = tempfile.mkstemp(
            dir=input_file_dir,
            prefix=f"{input_file_name}-labels-{dataset_type}-",
            suffix=input_file_ext,
        )
    updated_df = df.merge(label_df, on=const.ID, how=const.INNER)
    updated_df.to_csv(output_path, index=False)
    if delete_file:
        os.remove(df_path)
    return output_path
